/*
 * nlsop server code
 *
 * author: frank zirkelbach (frank.zirkelbach@physik.uni-augsburg.de)
 *
 * this program tries helping to understand the amorphous depuration
 * and recrystallization of SiCx while ion implantation at temperatures
 * below 400 degree celsius.
 * hopefully the program will simulate the stabilization of the
 * selforganizing lamella structure in the observed behaviour.
 *
 * refs: 
 *  - J. K. N. Lindner. Habil.Schrift, Universitaet Augsburg.
 *  - Maik Haeberlen. Diplomarbeit, Universitaet Augsburg.
 *
 * Copyright (C) 2004 Frank Zirkelbach
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include "nlsop.h"
#include "dfbapi.h"
#include "random.h"

#include "network.h"
#include "event.h"
#include "list.h"

#include "nlsop_general.h"

/* globals */
int *gi;
t_net *gnet;
t_event *gevent;
t_list *gc_list;
t_list *gg_list;
int alert;

/*
 * server specific stuff
 */

int usage(char *prog)
{
 puts("usage:");
 printf("%s <listen port>\n",prog);
 return 1;
}

int add_node(t_net *net,t_event *event,t_list *c_list,t_list *g_list) {

  int channel;
  unsigned char data;
  t_client client;
  int gui_chan;

  channel=network_manage_incoming(net);
  if(channel==N_E_ACCEPT) {
    printf("accept failed!\n");
    return -1;
  }
  if(channel==N_E_MAXC) {
    printf("maximum connections reached!\n");
    return -1;
  }
  printf("connection from %s port %d (ch: %d)\n",net->connection[channel].ip,
                                                 net->connection[channel].port,
                                                 channel);

  /* are you client or gui? */
  network_receive_chan(net,channel,&data,1);
  if(data==NLSOP_GUI) {
    gui_chan=channel;
    list_add_element(g_list,&gui_chan,sizeof(int));
    printf("node is a gui\n");
  }
  else if(data==NLSOP_CLIENT) {
    client.status=IDLE;
    client.channel=channel;
    list_add_element(c_list,&client,sizeof(t_client));
    printf("node is a client\n");
  }
  else {
    printf("not a client or gui - lets kick that ass out of here!\n");
    network_close(net,channel);
    return -1;
  }

  /* if we have a new node - care for it! */
  event_math(net->connection[channel].fd,event,READ,ADD);

  printf("\n");

  return 1;
}

int save_job(t_net *net,int channel,t_job *job,unsigned char dc) {

  char filename[128];
  int fd;
  int ret;
  d3_lattice d3l;
  info info;
  unsigned char data;

  printf("receiving data from client (%d)\n",job->size);
  data=DATA_OK;

  ret=network_receive_chan(net,channel,&d3l,sizeof(d3_lattice));
  network_send_chan(net,channel,&data,sizeof(unsigned char));
  printf("debug: got d3_lattice\n");

  ret=network_receive_chan(net,channel,&info,sizeof(info));
  network_send_chan(net,channel,&data,sizeof(unsigned char));
  printf("debug: got info\n");

  ret=network_receive_chan(net,channel,job->ac,job->size*sizeof(unsigned char));
  network_send_chan(net,channel,&data,sizeof(unsigned char));
  printf("debug: got ac\n");

  ret=network_receive_chan(net,channel,(unsigned char *)job->cc,
                           job->size*sizeof(int));
  network_send_chan(net,channel,&data,sizeof(unsigned char));
  printf("debug: got cc\n");

  ret=network_receive_chan(net,channel,(unsigned char *)&(job->step),
                           sizeof(int));
  network_send_chan(net,channel,&data,sizeof(unsigned char));
  printf("debug: got steps\n");

  if(dc!=DC_QUIT) {
    snprintf(filename,128,"./data/nlsop_b%f_c%f_s%f_ds%d_dr%f_-_%d_of_%d.save",
             job->info.b,job->info.c,job->info.s,
             job->info.diff_rate,job->info.dr_ac,
             job->step,job->info.steps);
    if((fd=open(filename,O_WRONLY|O_CREAT))<0) {
      printf("FATAL: unable to open file %s\n",filename);
      return -1;
    }

    if(write(fd,&d3l,sizeof(d3_lattice))<sizeof(d3_lattice)) {
      printf("FATAL: write of d3_lattice failed\n");
      return -1;
    }

    if(write(fd,&info,sizeof(info))<sizeof(info)) {
      printf("FATAL: write of info failed\n");
      return -1;
    }

    ret=write(fd,job->ac,job->size*sizeof(unsigned char));
    if(ret<job->size*sizeof(unsigned char)) {
      printf("FATAL: write of a/c states failed\n");
      return -1;
    }
   
    ret=write(fd,job->cc,job->size*sizeof(int));
    if(ret<job->size*sizeof(int)) {
      printf("FATAL: write of c.-conc. failed\n");
      return -1;
    }

    close(fd);
    
  }

  return 1;
}

int add_job(t_net *net,int chan,t_list *jl) {

  t_job job;

  job.channel=-1;
  job.status=IN_QUEUE;
  job.progress=0;

  network_receive_chan(net,chan,(unsigned char *)&(job.x),sizeof(int));
  network_receive_chan(net,chan,(unsigned char *)&(job.y),sizeof(int));
  network_receive_chan(net,chan,(unsigned char *)&(job.z),sizeof(int));
  network_receive_chan(net,chan,(unsigned char *)&(job.info),sizeof(info));

  job.size=job.x*job.y*job.z;

  job.ac=(unsigned char *)malloc(job.size*sizeof(unsigned char));
  if(job.ac==NULL) {
    printf("unable to malloc a/c memory\n");
    return -1;
  }

  job.cc=(int *)malloc(job.size*sizeof(int));
  if(job.cc==NULL) {
    printf("unable to malloc cc memory\n");
    return -1;
  }

  job.step=0;

  list_add_element(jl,&job,sizeof(t_job));
  
  printf("job added: b=%f | c=%f | s=%f ...\n",
         job.info.b,job.info.c,job.info.s);

  return 1;
}

int send_status(t_net *net,int chan,t_list *jl) {

  unsigned char data;
  int count;
  int i;

  data=GUI_INFO;

  count=list_count(jl);

  printf("sending job info\n");

  network_send_chan(net,chan,&data,sizeof(unsigned char));
  network_send_chan(net,chan,(unsigned char *)&count,sizeof(int));

  list_reset(jl);
  for(i=0;i<count;i++) {
    network_send_chan(net,chan,jl->current->data,sizeof(t_job));
    list_next(jl);
  }

  return 1;
}

int handle_node(t_net *net,t_event *event,
                t_list *c_list,t_list *g_list,t_list *job) {

  int i;
  unsigned char data;
  t_client *c;
  t_job *j;

  gi=&i;

  for(i=0;i<MAX_CONNECTIONS;i++) {
    if(FD_ISSET(net->connection[i].fd,&(event->rfds))) {

      alert=0;
      alarm(1);
      network_receive_chan(net,i,&data,1);
      alarm(0);

      if(alert==1) return -1;

      if(list_search_data(c_list,&i,sizeof(int))==L_SUCCESS) {
        /* it's a client */
	list_search_data(job,&i,sizeof(int));
        j=(t_job *)job->current->data;
        c=(t_client *)c_list->current-data;

        if(data==DC_END) {
          save_job(net,i,j,DC_END);
          /* reset client */
          c->status=IDLE;
          /* free job memory */
          free(j->ac);
          free(j->cc);
          /* delete job entry */
          list_del_current(job);
          printf("job ended, saved and removed from list.\n");
        }

        if(data==DC_OK) {
          save_job(net,i,j,DC_OK);
          /* inc progress state */
          j->progress+=1;
          printf("job at next level, saved.\n");
        }

        if(data==DC_QUIT) {
          save_job(net,i,j,DC_QUIT);
          /* network disconnect */
          event_math(net->connection[i].fd,event,READ,REMOVE);
          network_close(net,i);
          /* del from client list */
          list_del_current(c_list);
          /* change job state */
          j->status=IN_QUEUE;
          printf("client terminating, job queued, client removed.\n");
        }
      }

      else if(list_search_data(g_list,&i,sizeof(int))==L_SUCCESS) {
        /* its a gui */
        if(data==GUI_ADDJOB) add_job(net,i,job);

        else if(data==GUI_INFO) send_status(net,i,job);

        else if(data==GUI_QUIT) {
          printf("disconnecting gui on channel %d\n",i);
          event_math(net->connection[i].fd,event,READ,REMOVE);
          network_close(net,i);
          list_del_current(g_list);
        }

        else {
          printf("unknown gui command\n");
          return -1;
        }
      }

      else {
        printf("this chan is not in client or gui list! i disconnect now!\n");
        event_math(net->connection[i].fd,event,READ,REMOVE);
        network_close(net,i);
      }
    }
  }

  printf("\n");
   
  return 1;
}

int distribute_jobs(t_event *event,void *allineed) {

  t_net *net;
  t_list *c_list,*g_list,*job;
  int count_j,count_c,min;
  t_job *j;
  t_client *c;
  unsigned char data;
  d3_lattice d3l;
  unsigned int addr[4];

  memcpy(addr,allineed,4*sizeof(unsigned int));

  net=(t_net *)addr[0];
  c_list=(t_list *)addr[1];
  g_list=(t_list *)addr[2];
  job=(t_list *)addr[3];

  list_reset(job);
  list_reset(c_list);

  count_j=0;
  count_c=0;

  if((c_list->current==NULL)||(job->current==NULL)) return 2;

  j=(t_job *)job->current->data;
  c=(t_client *)c_list->current->data;
  if(j->status==IN_QUEUE) count_j++;
  if(c->status==IDLE) count_c++;

  while(list_next(job)!=L_NO_NEXT_ELEMENT) {
    j=(t_job *)job->current->data;
    if(j->status==IN_QUEUE) count_j++;
  }
  while(list_next(c_list)!=L_NO_NEXT_ELEMENT) {
    c=(t_client *)c_list->current->data;
    if(c->status==IDLE) count_c++;
  }
 
  min=(count_c<count_j)?count_c:count_j;

  if(min!=0) {
    printf("d: distributing jobs ...\n");
    printf("%d queued jobs, %d idle clients\n\n",count_j,count_c);
  }

  list_reset(job);
  list_reset(c_list);

  while(min) {
    j=(t_job *)job->current->data;
    c=(t_client *)c_list->current->data;
    while(c->status!=IDLE) {
      list_next(c_list);
      c=(t_client *)c_list->current->data;
    }
    while(j->status!=IN_QUEUE) {
      list_next(job);
      j=(t_job *)job->current->data;
    }

    /* direct current job to current client */
    if(j->step==0) data=NLSOP_NJOB;
    else data=NLSOP_CJOB;

    c->status=WORK;
    j->channel=c->channel;
    j->status=IN_WORK;

    d3l.max_x=j->x;
    d3l.max_y=j->y;
    d3l.max_z=j->z;

    network_send_chan(net,c->channel,&data,sizeof(unsigned char));
    network_send_chan(net,c->channel,(unsigned char *)&d3l,sizeof(d3_lattice));
    network_send_chan(net,c->channel,(unsigned char *)&(j->info),sizeof(info));

    if(data==NLSOP_CJOB) {
      network_send_chan(net,c->channel,j->ac,j->size*sizeof(unsigned char));
      network_receive_chan(net,c->channel,&data,sizeof(unsigned char));
      network_send_chan(net,c->channel,(unsigned char *)&(j->cc),
                        j->size*sizeof(int));
      network_receive_chan(net,c->channel,&data,sizeof(unsigned char));
      network_send_chan(net,c->channel,(unsigned char *)&(j->step),sizeof(int));
      network_receive_chan(net,c->channel,&data,sizeof(unsigned char));
    }

    --min;
    list_next(c_list);
    list_next(job);
  }

  return 1;
}

int parse_incoming(t_event *event,void *allineed) {

  t_net *net;
  t_list *c_list,*g_list,*job;
  unsigned int addr[4];

  memcpy(addr,allineed,4*sizeof(unsigned int));

  net=(t_net *)addr[0];
  c_list=(t_list *)addr[1];
  g_list=(t_list *)addr[2];
  job=(t_list *)addr[3];

  /* decide what to do */
  if(FD_ISSET(net->l_fd,&(event->rfds))) {
    /* new node */
    printf("new node ...\n");
    add_node(net,event,c_list,g_list);
  }
  else {
    /* client/gui interaction */
    printf("node interaction ...\n");
    handle_node(net,event,c_list,g_list,job);
  }
    
  return 1;
}

void destroy_it(int signum) {

  printf("connection to client (ch %d) fucked up!\n",*gi);
  event_math(gnet->connection[*gi].fd,gevent,READ,REMOVE);
  network_close(gnet,*gi);
  if(list_search_data(gc_list,gi,sizeof(int))==L_SUCCESS) {
    list_del_current(gc_list);
    printf("removed client from list\n");
  }
  if(list_search_data(gg_list,gi,sizeof(int))==L_SUCCESS) {
    list_del_current(gg_list);
    printf("removed gui from list\n");
  }

  alert=1;
  alarm(0);

}

/*
 * main program
 */

int main(int argc,char **argv)
{

  int port;
  t_net net;
  t_event event;
  t_list c_list;
  t_list g_list;
  t_list job;
  void *allyouneed;
  unsigned int addr[4];

  gnet=&net;
  gevent=&event;
  gc_list=&c_list;
  gg_list=&g_list;

  /* tzzz ... */
  allyouneed=(void *)addr;
  addr[0]=(unsigned int)&net;
  addr[1]=(unsigned int)&c_list;
  addr[2]=(unsigned int)&g_list;
  addr[3]=(unsigned int)&job;
  
  /* default values */
  port=1025;

  /* parse argv */
  if(argc==2) port=atoi(argv[1]);

  /* event init */
  event_init(&event,1);
  /* 10 sec event timeout - distributing jobs */
  event_set_timeout(&event,10,0);

  /* list init */
  list_init(&c_list,1);
  list_init(&g_list,1);
  list_init(&job,1);

  /* connect to server */
  network_init(&net,1);
  network_set_listen_port(&net,port);
  if(network_listen(&net)!=N_SUCCESS) {
    printf("unable to listen on port %d, aborting!\n",port);
    return -1;
  }

  /* install sighandler */
  signal(SIGALRM,destroy_it);

  /* wait for events :) */
  event_math(net.l_fd,&event,READ,ADD);
  printf("\nNLSOP_SERVER started!\n\n");
  event_start(&event,allyouneed,parse_incoming,distribute_jobs);

  return 1;
}

